__author__ = "geduldig"
__date__ = "November 30, 2014"
__license__ = "MIT"


import logging
import json


class TwitterError(Exception):

    """Base class for Twitter exceptions"""
    pass


class TwitterConnectionError(TwitterError):

    """Raised when the connection needs to be re-established"""

    def __init__(self, value):
        super().__init__(value)
        logging.warning('%s %s' % (type(value), value))


class TwitterRequestError(TwitterError):

    """Raised when request fails"""

    def __init__(self, status_code, msg=None):
        if msg is None:
            if status_code >= 500:
                msg = 'Twitter internal error (you may re-try)'
            else:
                msg = 'Twitter request failed'
        logging.info('Status code %d: %s' % (status_code, msg))
        super().__init__(msg)
        self.status_code = status_code
        self.msg = msg

    def __str__(self):
        return '%s (%d): %s' % (self.args, self.status_code, self.msg)


    def __iter__(self):
        try:
            msg = json.loads(self.msg)
            if 'errors' in msg:
                for error in msg['errors']:
                    yield error['message']
            elif 'detail' in msg:
                yield msg['detail']
            else:
                yield self.msg
        except:
            yield self.msg


__author__ = "geduldig"
__date__ = "June 8, 2013"
__license__ = "MIT"


from requests.exceptions import ConnectionError, ReadTimeout, SSLError
from requests.packages.urllib3.exceptions import ReadTimeoutError, ProtocolError
from .TwitterError import *
import requests
import time


class TwitterPager(object):

    """Continuous (stream-like) pagination of response from Twitter REST API resource.
    In addition to Public API endpoints, supports Premium Search API.

    :param api: An authenticated TwitterAPI object
    :param resource: String with the resource path (ex. search/tweets)
    :param params: Dictionary of resource parameters
    """

    def __init__(self, api, resource, params=None):
        self.api = api
        self.resource = resource
        self.params = params

    def get_iterator(self, wait=5, new_tweets=False):
        """Iterate response from Twitter REST API resource. Resource is called
        in a loop to retrieve consecutive pages of results.

        :param wait: Floating point number (default=5) of seconds wait between requests.
                     Depending on the resource, appropriate values are 5 or 60 seconds.
        :param new_tweets: Boolean determining the search direction.
                           False (default) retrieves old results.
                           True retrieves current results.
        :param version: override default API version or None (default)

        :returns: JSON objects containing statuses, errors or other return info.
        """
        elapsed = 0
        while True:
            try:
                # REQUEST ONE PAGE OF RESULTS...
                start = time.time()
                r = self.api.request(self.resource, self.params)
                it = r.get_iterator()
                if new_tweets:
                    it = reversed(list(it))

                # YIELD FOR EACH ITEM IN THE PAGE...
                item_count = 0
                id = None
                for item in it:
                    item_count += 1
                    if type(item) is dict:
                        if 'id' in item:
                            id = item['id']
                        if 'code' in item:
                            if item['code'] in [130, 131]:
                                # Twitter service error
                                raise TwitterConnectionError(item)
                    yield item

                json = r.json()

                # CHECK FOR NEXT PAGE OR BAIL...
                if self.api.version == '1.1':
                    # if a cursor is present, use it to get next page
                    # otherwise, use id to get next page
                    is_premium_search = self.params and 'query' in self.params and self.api.version == '1.1'
                    cursor = -1
                    if new_tweets and 'previous_cursor' in json:
                        cursor = json['previous_cursor']
                        cursor_param = 'cursor'
                    elif not new_tweets:
                        if 'next_cursor' in json:
                            cursor = json['next_cursor']
                            cursor_param = 'cursor'
                        elif 'next' in json:
                            # 'next' is used by Premium Search (OLD searches only)
                            cursor = json['next']
                            cursor_param = 'next'

                    # bail when no more results
                    if cursor == 0:
                        break
                    elif cursor == -1 and is_premium_search:
                        break
                    elif not new_tweets and item_count == 0:
                        break
                else: # VERSION 2
                    meta = json['meta']

                # SLEEP...
                elapsed = time.time() - start
                pause = wait - elapsed if elapsed < wait else 0
                time.sleep(pause)

                # SETUP REQUEST FOR NEXT PAGE...
                if self.api.version == '1.1':
                    # get a page with cursor if present, or with id if not
                    # a Premium search (i.e. 'query' is not a parameter)
                    if cursor != -1:
                        self.params[cursor_param] = cursor
                    elif id is not None and not is_premium_search:
                        if new_tweets:
                            self.params['since_id'] = str(id)
                        else:
                            self.params['max_id'] = str(id - 1)
                    else:
                        continue
                else: # VERSION 2
                    if new_tweets:
                        self.params['since_id'] = meta['newest_id']
                    else:
                        self.params['next_token'] = meta['next_token']

            except TwitterRequestError as e:
                if e.status_code < 500:
                    raise
                continue
            except TwitterConnectionError:
                continue


"""
	Constants For All Twitter Endpoints
	-----------------------------------
	
	Version 1.1, Streaming API and REST API.
	
	URLs for each endpoint are composed of the following pieces:
		PROTOCOL://{subdomain}.DOMAIN/VERSION/{resource}?{parameters}
"""


__author__ = "geduldig"
__date__ = "February 3, 2012"
__license__ = "MIT"


PROTOCOL = 'https'
DOMAIN = 'twitter.com'

VERSION = '1.1'
CURATOR_VERSION = 'broadcast/1'
ADS_VERSION = '2'

ENDPOINTS = {
    # resource:                                             (method,   subdomain)

    # STREAMING API

    'statuses/filter':                                      ('POST',   'stream'),
    'statuses/firehose':                                    ('GET',    'stream'),
    'statuses/sample':                                      ('GET',    'stream'),
    'site':                                                 ('GET',    'sitestream'),
    'user':                                                 ('GET',    'userstream'),

    # PUBLIC API

    'account/remove_profile_banner':                        ('POST',   'api'),
    'account/settings':                                     ('GET',    'api'),
    'account/update_delivery_device':                       ('POST',   'api'),
    'account/update_profile':                               ('POST',   'api'),
    'account/update_profile_background_image':              ('POST',   'api'),
    'account/update_profile_banner':                        ('POST',   'api'),
    'account/update_profile_colors':                        ('POST',   'api'),
    'account/update_profile_image':                         ('POST',   'api'),
    'account/verify_credentials':                           ('GET',    'api'),
        
    'application/rate_limit_status':                        ('GET',    'api'),

    'blocks/create':                                        ('POST',   'api'),
    'blocks/destroy':                                       ('POST',   'api'),
    'blocks/ids':                                           ('GET',    'api'),
    'blocks/list':                                          ('GET',    'api'),

    'direct_messages':                                      ('GET',    'api'), # deprecated
    'direct_messages/destroy':                              ('POST',   'api'), # deprecated
    'direct_messages/events/destroy':                       ('DELETE', 'api'),
    'direct_messages/events/new':                           ('POST',   'api'),
    'direct_messages/events/list':                          ('GET',    'api'),
    'direct_messages/events/show':                          ('GET',    'api'),    
    'direct_messages/indicate_typing':                      ('POST',   'api'),
    'direct_messages/new':                                  ('POST',   'api'), # deprecated
    'direct_messages/sent':                                 ('GET',    'api'), # deprecated
    'direct_messages/show':                                 ('GET',    'api'), # deprecated
    'direct_messages/welcome_messages/new':                 ('POST',   'api'),
    'direct_messages/welcome_messages/list':                ('GET',    'api'),
    'direct_messages/welcome_messages/show':                ('GET',    'api'),
    'direct_messages/welcome_messages/destroy':             ('DELETE', 'api'),    
    'direct_messages/welcome_messages/rules/new':           ('POST',   'api'),
    'direct_messages/welcome_messages/rules/list':          ('GET',    'api'),
    'direct_messages/welcome_messages/rules/show':          ('GET',    'api'),    
    'direct_messages/welcome_messages/rules/destroy':       ('DELETE', 'api'),        

    'favorites/create':                                     ('POST',   'api'),
    'favorites/destroy':                                    ('POST',   'api'),
    'favorites/list':                                       ('GET',    'api'),

    'followers/ids':                                        ('GET',    'api'),
    'followers/list':                                       ('GET',    'api'),

    'friends/ids':                                          ('GET',    'api'),
    'friends/list':                                         ('GET',    'api'),

    'friendships/create':                                   ('POST',   'api'),
    'friendships/destroy':                                  ('POST',   'api'),
    'friendships/incoming':                                 ('GET',    'api'),
    'friendships/lookup':                                   ('GET',    'api'),
    'friendships/no_retweets/ids':                          ('GET',    'api'),
    'friendships/outgoing':                                 ('GET',    'api'),
    'friendships/show':                                     ('GET',    'api'),
    'friendships/update':                                   ('POST',   'api'),

    'lists/create':                                         ('POST',   'api'),
    'lists/destroy':                                        ('POST',   'api'),
    'lists/list':                                           ('GET',    'api'),
    'lists/members':                                        ('GET',    'api'),
    'lists/members/create':                                 ('POST',   'api'),
    'lists/members/create_all':                             ('POST',   'api'),
    'lists/members/destroy':                                ('POST',   'api'),
    'lists/members/destroy_all':                            ('POST',   'api'),
    'lists/members/show':                                   ('GET',    'api'),
    'lists/memberships':                                    ('GET',    'api'),
    'lists/ownerships':                                     ('GET',    'api'),
    'lists/show':                                           ('GET',    'api'),
    'lists/statuses':                                       ('GET',    'api'),
    'lists/subscribers':                                    ('GET',    'api'),
    'lists/subscribers/create':                             ('POST',   'api'),
    'lists/subscribers/destroy':                            ('POST',   'api'),
    'lists/subscribers/show':                               ('GET',    'api'),
    'lists/subscriptions':                                  ('GET',    'api'),
    'lists/update':                                         ('POST',   'api'),

    'media/metadata/create':                                ('POST',   'upload'),
    'media/upload':                                         ('POST',   'upload'),

    'mutes/users/create':                                   ('POST',   'api'),
    'mutes/users/destroy':                                  ('POST',   'api'),
    'mutes/users/ids':                                      ('GET',    'api'),
    'mutes/users/list':                                     ('GET',    'api'),

    'geo/id/:PARAM':                                        ('GET',    'api'), # PLACE_ID
    'geo/place':                                            ('POST',   'api'),
    'geo/reverse_geocode':                                  ('GET',    'api'),
    'geo/search':                                           ('GET',    'api'),
    'geo/similar_places':                                   ('GET',    'api'),

    'help/configuration':                                   ('GET',    'api'),
    'help/languages':                                       ('GET',    'api'),
    'help/privacy':                                         ('GET',    'api'),
    'help/tos':                                             ('GET',    'api'),

    'saved_searches/create':                                ('POST',   'api'),
    'saved_searches/destroy/:PARAM':                        ('POST',   'api'), # ID
    'saved_searches/list':                                  ('GET',    'api'),
    'saved_searches/show/:PARAM':                           ('GET',    'api'), # ID

    'search/tweets':                                        ('GET',    'api'),

    'statuses/destroy/:PARAM':                              ('POST',   'api'), # ID
    'statuses/home_timeline':                               ('GET',    'api'),
    'statuses/lookup':                                      ('GET',    'api'),
    'statuses/mentions_timeline':                           ('GET',    'api'),
    'statuses/oembed':                                      ('GET',    'api'),
    'statuses/retweet/:PARAM':                              ('POST',   'api'), # ID
    'statuses/retweeters/ids':                              ('GET',    'api'),
    'statuses/retweets/:PARAM':                             ('GET',    'api'), # ID
    'statuses/retweets_of_me':                              ('GET',    'api'),
    'statuses/show/:PARAM':                                 ('GET',    'api'), # ID
    'statuses/unretweet/:PARAM':                            ('POST',   'api'), # ID
    'statuses/user_timeline':                               ('GET',    'api'),
    'statuses/update':                                      ('POST',   'api'),
    'statuses/update_with_media':                           ('POST',   'api'), # deprecated

    'trends/available':                                     ('GET',    'api'),
    'trends/closest':                                       ('GET',    'api'),
    'trends/place':                                         ('GET',    'api'),

    'users/contributees':                                   ('GET',    'api'),
    'users/contributors':                                   ('GET',    'api'),
    'users/lookup':                                         ('POST',   'api'),
    'users/profile_banner':                                 ('GET'     'api'),
    'users/report_spam':                                    ('POST',   'api'),
    'users/search':                                         ('GET',    'api'),
    'users/show':                                           ('GET',    'api'),
    'users/suggestions':                                    ('GET',    'api'),
    'users/suggestions/:PARAM':                             ('GET',    'api'), # SLUG
    'users/suggestions/:PARAM/members':                     ('GET',    'api'), # SLUG

    # COLLECTIONS API

    'collections/create':                                   ('POST',   'api'),
    'collections/destroy':                                  ('POST',   'api'),
    'collections/entries':                                  ('GET',    'api'),
    'collections/entries/add':                              ('POST',   'api'),
    'collections/entries/curate':                           ('POST',   'api'),
    'collections/entries/move':                             ('POST',   'api'),
    'collections/entries/remove':                           ('POST',   'api'),
    'collections/list':                                     ('GET',    'api'),
    'collections/show':                                     ('GET',    'api'),
    'collections/update':                                   ('POST',   'api'),

    # CURATOR API

    'collections/:PARAM/content':                           ('GET',    'curator'), # ID
    'projects':                                             ('GET',    'curator'),
    'projects/:PARAM':                                      ('GET',    'curator'), # ID
    'streams/:PARAM/content':                               ('GET',    'curator'), # ID
    'streams/:PARAM/metrics':                               ('GET',    'curator'), # ID
    'streams/:PARAM/trendline':                             ('GET',    'curator'), # ID
    'streams/compare':                                      ('GET',    'curator'),
    'streams/compare_to_target':                            ('GET',    'curator'),

    # ADS API (not tested!!)

    'accounts/:PARAM/auction_insights':                     ('GET',    'ads-api'), # ACCOUNT ID
    'stats/accounts/:PARAM':                                ('GET',    'ads-api'), # ACCOUNT ID
    'stats/accounts/:PARAM/reach/funding_instruments':      ('GET',    'ads-api'), # ACCOUNT ID
    'stats/jobs/accounts/:PARAM':                           ('GET',    'ads-api'), # ACCOUNT ID
    'stats/jobs/accounts/:PARAM':                           ('POST',   'ads-api'), # ACCOUNT ID
    'stats/jobs/accounts/:PARAM/:PARAM':                    ('DELETE', 'ads-api'), # ACCOUNT ID, JOB ID
    'stats/jobs/summaries':                                 ('GET',    'ads-api'),
    
    # ACCOUNT ACTIVITY WEBHOOK API

    'account_activity/all/:PARAM/subscriptions':            ('POST',   'api'), # ENVIRONMENT NAME
    'account_activity/all/:PARAM/subscriptions':            ('DELETE', 'api'), # ENVIRONMENT NAME                            
    'account_activity/all/:PARAM/subscriptions/all':        ('GET',    'api'), # ENVIRONMENT NAME
    'account_activity/all/:PARAM/subscriptions/all/list':   ('GET',    'api'), # ENVIRONMENT NAME
    'account_activity/all/:PARAM/subscriptions/list':       ('GET',    'api'), # ENVIRONMENT NAME
    'account_activity/all/:PARAM/webhooks':                 ('POST',   'api'), # ENVIRONMENT NAME    
    'account_activity/all/:PARAM/webhooks/:PARAM':          ('DELETE', 'api'), # ENVIRONMENT NAME, WEBHOOK ID
    'account_activity/all/count':                           ('GET',    'api'), 
    'account_activity/all/webhooks':                        ('GET',    'api'),   
    'account_activity/webhooks':                            ('POST',   'api'),    
    'account_activity/webhooks/:PARAM':                     ('DELETE', 'api'), # WEBHOOK ID
    'account_activity/webhooks/:PARAM/subscriptions':       ('POST',   'api'), # ENVIRONMENT NAME   
    'account_activity/webhooks/:PARAM/subscriptions/list':  ('GET',    'api'), # ENVIRONMENT NAME

    # PREMIUM SEARCH API

    'tweets/search/30day/:PARAM':                           ('GET',    'api'), # LABEL
    'tweets/search/30day/:PARAM/counts':                    ('GET',    'api'), # LABEL
    'tweets/search/fullarchive/:PARAM':                     ('GET',    'api'), # LABEL
    'tweets/search/fullarchive/:PARAM/counts':              ('GET',    'api'), # LABEL
    
    # LABS API (BETAS) WILL NEED APPLICATION APPROVAL

    'labs/1/tweets/metrics/private':                        ('GET',    'api'),
    'labs/2/tweets/:PARAM':                                 ('GET',    'api'), # TWEET ID
    'labs/2/tweets':                                        ('GET',    'api'),
    'labs/2/tweets/search':                                 ('GET',    'api'),
    'labs/2/tweets/:PARAM/hidden':                          ('PUT',    'api'), # TWEET ID
    'labs/2/users/:PARAM':                                  ('GET',    'api'), # USER ID
    'labs/2/users':                                         ('GET',    'api'),          
    
    # API V2 EARLY ACCESS

    'tweets':                                               ('GET',    'api'),
    'tweets/:PARAM/hidden':                                 ('PUT',    'api'), # ID
    'tweets/:PARAM':                                        ('GET',    'api'), # ID
    'tweets/sample/stream':                                 ('GET',    'api'),
    'tweets/search/recent':                                 ('GET',    'api'),
    'tweets/search/stream':                                 ('GET',    'api'),
    'tweets/search/stream/rules':                           ('POST',   'api'), # use method_override for 'GET'
    'users':                                                ('GET',    'api'),
    'users/:PARAM':                                         ('GET',    'api'), # ID
    'users/by':                                             ('GET',    'api'),
    'users/by/username/:PARAM':                             ('GET',    'api'), # USERNAME
}
